<?php

/*	LibFox™ - Simple library for rapid development of web applications.
	Copyright © 2007,2008 LinFox Serviços em Informatica LTDA.

	This file is part of LibFox.

	LibFox is free software: you can redistribute it and/or modify
	it under the terms of the GNU General Public License as published by
	the Free Software Foundation, either version 3 of the License, or
	(at your option) any later version.

	LibFox is distributed in the hope that it will be useful,
	but WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
	GNU General Public License for more details.

	You should have received a copy of the GNU General Public License
	along with LibFox.  If not, see <http://www.gnu.org/licenses/>. */

/**
 * Data Access Object base class file.
 * 
 * This file contains the abstract dao class.
 * @author Alwin Garside <alwin@linfox.com.br>
 * @copyright Copyright (c) 2007,2008 LinFox Serviços de Informatica LTDA.
 * @license http://www.gnu.org/licenses/gpl.html GNU General Public License
 * @version 0.1
 * @package libfox
 */

/**
 * Include dependencies.
 */
require_once 'lib/model.php';

/**
 * Abstract base class for Data Access Objects
 * 
 * This is used to retrieve and modify entries
 * related to models in a database.
 * 
 * @package libfox
 * @abstract
 */
class dao
{
	/**
	 * Primary key
	 * 
	 * A comma seperated list of columns that serve as the tables' primary
	 * key.
	 * @var string
	 */
	public $key = 'id';

	/**
	 * Foreign keys
	 * 
	 * A comma seperated list of fields that serve as foreign keys.
	 * This will be used to match results according to foreign keys. The idea
	 * is that these fields must be the same in all tables where they are used.
	 *
	 * @var string
	 */
	public $foreign_keys = NULL;

	/**
	 * Other keys
	 *
	 * Comma-seperated list of other keys to match the model by
	 * @var string
	 */
	public $other_keys = NULL;

	/**
	 * Table name
	 *
	 * @var string
	 */
	protected $table = 'models';

	/**
	 * Model class
	 * 
	 * Name of the class to create new models from.
	 * @var string
	 */
	protected $model = 'model';

	/**
	 * Fields to select when using fetch()
	 * 
	 * A comma seperated list of fields. They must exist in this dao's table
	 * and will be used by fetch(). If you wish to fetch fields from other
	 * tables, use $fields_rel and $joins.
	 * @var string
	 */
	protected $fields = NULL;

	/**
	 * Fields to select using SUM()
	 *
	 * @var string
	 */
	protected $fields_sum = NULL;

	/**
	 * Fields to select using COUNT()
	 *
	 * @var string
	 */
	protected $fields_count = NULL;

	/**
	 * Fields to select when using fetch_all()
	 * 
	 * @var string
	 */
	protected $fields_all = NULL;

	/**
	 * Fields to fetch from related tables with fetch()
	 * [$name => [$table => fields]]
	 * 
	 * @var array
	 */
	protected $fields_rel = NULL;

	/**
	 * Fields to fetch from related tables with fetch()
	 * [$name => [$table => fields]]
	 * 
	 * @var array
	 */
	protected $fields_rel_all = NULL;

	/**
	 * Additional SELECT SQL code
	 *
	 * @var string
	 */
	protected $select_sql = NULL;

	/**
	 * Joins
	 * 
	 * Array with joins that have to be made to get the related fields defined
	 * in $fields_rel.
	 * @example ['table1.field' => 'table2.field']
	 * @var string
	 */
	protected $joins = NULL;

	/**
	 * Left joins
	 * 
	 * Array with left joins that have to be made to get the related fields defined
	 * in $fields_rel.
	 * @example ['table1.field' => 'table2.field']
	 * @var string
	 */
	protected $left_joins = NULL;

	/**
	 * Right joins
	 * 
	 * Array with right joins that have to be made to get the related fields defined
	 * in $fields_rel.
	 * @example ['table1.field' => 'table2.field']
	 * @var string
	 */
	protected $right_joins = NULL;

	/**
	 * Additional SQL to append to JOIN
	 *
	 * @var string
	 */
	protected $joins_sql = false;

	/**
	 * Additional WHERE statement
	 * 
	 * @var string
	 */
	protected $where = false;

	/**
	 * Fields to search through when using fetch_all_like()
	 * 
	 * These fields will be searched through using LIKE "%FOO%"
	 * Like this: 'table.foo,table.bar'
	 * @var mixed
	 */
	protected $search_fields = false;

	/**
	 * Fields to search through when using fetch_all_like()
	 * 
	 * These fields will be searched through using LIKE "FOO%"
	 * @example table.foo,table.bar
	 * @var mixed
	 */
	protected $search_fields_l = false;

	/**
	 * Fields to search through when using fetch_all_like()
	 * 
	 * These fields will be searched through using LIKE "%FOO"
	 * @example table.foo,table.bar
	 * @var mixed
	 */
	protected $search_fields_r = false;

	/**
	 * Fields to order fetch_all() by
	 * comma seperated.
	 * 
	 * @var string
	 */
	public $order_by = false;

	/**
	 * Fields to group queries by,
	 * comma seperated.
	 * 
	 * @var mixed
	 */
	public $group_by = false;

	/**
	 * List of tables with relations to delete.
	 * Example: 'table1,table2'
	 *
	 * @var unknown_type
	 */
	protected $delete_rel = false;

	public $limit = false;

	public $offset = false;

	/**
	 * Database instance.
	 *
	 * @var PDO
	 */
	protected $db;

	/**
	 * dao's constructor
	 * 
	 * Grabs the global database instance using db::get_instance().
	 */
	public function __construct($table = false, $key = false, $model = false)
	{
		if ($table) $this->table = $table;
		if ($key) $this->key = $key;
		if ($model) $this->model = $model;
		$this->db = db::get_instance();
	}
	
	/**
	 * Fetch a model from the database.
	 *
	 * @param string $id
	 * @return model Returns a model or false 
	 */
	public function fetch($id = 1)
	{
		$sql = '
			SELECT '.$this->sql_fetch_fields().'
			FROM `'.$this->table.'`'
			.$this->sql_joins()
			.$this->sql_where($id)
			.$this->sql_group_by()
			.$this->sql_order_by();
		$result = $this->db->query($sql);

		if (!$model = $result->fetchObject($this->model)) {
			throw new Exception(lang::not_found.': '.lang::in.' '.$this->table.'.'.$this->key);
			return false;
		}
		$result->closeCursor();
		$model->set_dao($this); 
		return $model;
	}

	/**
	 * Fetch an array with models from the database.
	 *
	 * @param mixed $id
	 * @param string $keys example 'table.foo,table.bar'
	 * @return array An array with the models.
	 */
	public function fetch_all($search_model = NULL)
	{
		$models = array();
		$sql  = 'SELECT '.$this->sql_fetch_all_fields().'
			FROM `'.$this->table.'`'
			.$this->sql_joins()
			.$this->sql_where($search_model)
			.$this->sql_group_by()
			.$this->sql_order_by()
			.$this->sql_limit();

		$result = $this->db->query($sql);
		$result->setFetchMode(PDO::FETCH_CLASS,$this->model);
		return $result->fetchAll();
	}

	/**
	 * Fetch an array with models from the database.
	 * 
	 * @param string $search
	 * @return array
	 */
	public function fetch_all_like($search_model = false, $search = false)
	{
		$models = array();
		$sql  = 'SELECT '.$this->sql_fetch_all_fields().'
			FROM `'.$this->table.'`'
			.$this->sql_joins()
			.$this->sql_where_like($search_model, $search)
			.$this->sql_group_by()
			.$this->sql_order_by()
			.$this->sql_limit();

		$result = $this->db->query($sql);
		$result->setFetchMode(PDO::FETCH_CLASS,$this->model);
		return $result->fetchAll();
	}

	public function count($field, $id = false, $key = false)
	{
		$sql = '
			SELECT COUNT(`'.$field.'`)
			FROM `'.$this->table.'`'
			.$this->sql_where($id,$key);
		$result = $this->db->query($sql);
		if (!$count = $result->fetchColumn()) {
			throw new Exception(lang::not_found.': '.$id.' '.lang::in.' '.$this->table.'.'.$key);
			return false;
		}
		$result->closeCursor();
		return $count;
	}
	
	/**
	 * Fetch an empty model structure.
	 *
	 * @return model
	 */
	public function fetch_model()
	{
		$model = new $this->model();
		$sql = 'SHOW FIELDS FROM `'.$this->table.'`';
		$result = $this->db->query($sql);
		$result->setFetchMode(PDO::FETCH_CLASS,$this->model);
		$columns = $result->fetchAll();
		foreach ($columns as $column) {
			$field = $column->Field;
			$model->$field = '';
		}
		$model->set_dao($this);
		return $model;
	}

	/**
	 * Create a new entry in the table.
	 * 
	 * Returns a model with the key if the created row.
	 *
	 * @return model
	 */
	public function create()
	{
		$key = $this->key;
		$sql = 'INSERT INTO `'.$this->table.'` VALUES()';
		$this->db->query($sql);
		$model = new $this->model;
		$model->$key = $this->db->lastInsertId();
		$model->set_dao($this);
		return $model;
	}

	/**
	 * Insert's a model into the database.
	 *
	 * @param model $model
	 * @return boolean
	 */
	public function insert(&$model)
	{
		$key = $this->key;
		if (get_class($model) != $this->model) {
			throw new Exception('object_of_incorrect_class');
			return false;
		}
		$struct = $this->fetch_model();
		$sql_insert = 'INSERT INTO `'.$this->table.'` SET ';
		foreach ($struct as $field => $value) {
			if ($model->$field) {
				$sql_insert .= ' `'.$field.'` = "'.$model->$field.'",';
			}
		}
		$sql_insert = substr($sql_insert, 0, -1);

		if (!$this->db->query($sql_insert)) {
			throw new Exception($this->model.'_not_inserted');
			return false;
		}
		$model->$key = $this->db->lastInsertId();
		return true;
	}

	/**
	 * Update model in database.
	 *
	 * @param model $model
	 * @return boolean
	 */
	public function update(&$model)
	{
		if (get_class($model) != $this->model) {
			throw new Exception('object_of_incorrect_class');
			return false;
		}
		$key = $this->key;
		$model_db = $this->fetch_model();
		$sql_update = 'UPDATE `'.$this->table.'` SET';
		foreach ($model_db as $field => $value) {
			if ($field != $this->key && isset($model->$field) && $field != 'password') {
				$sql_update .= ' `'.$field.'` = "'.addslashes($model->$field).'",';
			}
		}
		$sql_update = substr($sql_update, 0, -1);
		$sql_update .= ' WHERE `'.$this->key.'` = "'.$model->$key.'" ';

		if (!$this->db->query($sql_update)) {
			throw new Exception($this->model.'_not_updated');
			return false;
		}
		return true;
	}

	/**
	 * Delete row from database.
	 *
	 * @param mixed $id
	 * @return boolean
	 */
	public function delete($id)
	{
		if (!$id) {
			throw new Exception('no_delete_key');
		}
		$sql_delete = '
			DELETE FROM `'.$this->table.'`
			WHERE `'.$this->key.'` = "'.addslashes($id).'"';
		if (!$this->db->query($sql_delete)) return false;

		if ($this->delete_rel) {
			$rels = explode(',',$this->delete_rel);
			foreach ($rels as $rel) {
				$sql_delete = '
					DELETE FROM `'.$rel.'`
					WHERE `'.$this->key.'` = "'.addslashes($id).'"';
				if (!$this->db->query($sql_delete)) return false;
			}
		}

		return true;
	}

	/**
	 * Replaces \r in a string with \n
	 *
	 * @param string $string
	 * @return string
	 */
	protected function str_clean($string)
	{
		return str_replace('\r','\n',$string);
	}

	protected function sql_fetch_fields()
	{
		$result = '';
		if ($this->fields) {
			$fields = explode(',',$this->fields);
			foreach ($this->fields as $field) {
				$result .= ' `'.$this->table.'`.`'.$field.'`,';
			}
		} else {
			$result .= ' `'.$this->table.'`.*,';
		}
		if ($this->fields_rel) {
			foreach ($this->fields_rel as $field_name => $fields) {
				foreach ($fields as $table => $field) {
					$result .= ' `'.$table.'`.`'.$field.'` AS `'.$field_name.'`,';
				}
			}
		}
		if ($this->fields_rel_all) {
			foreach ($this->fields_rel_all as $field_name => $fields) {
				foreach ($fields as $table => $field) {
					$result .= ' `'.$table.'`.`'.$field.'` AS `'.$field_name.'`,';
				}
			}
		}		
		if ($this->select_sql) $result .= ' '.$this->select_sql;
		else $result = substr($result, 0, -1).' ';
		return $result;
	}

	protected function sql_fetch_all_fields()
	{
		$result = '';
		if ($this->fields_all) {
			$fields = explode(',',$this->fields_all);
			foreach ($fields as $field) {
				$result .= ' `'.$this->table.'`.`'.$field.'`,';
			}
		} else {
			$result .= ' `'.$this->table.'`.*,';
		}
		if ($this->fields_sum) {
			foreach ($this->fields_sum as $field_name => $fields) {
				foreach ($fields as $table => $field) {
					$result .= ' SUM(`'.$table.'`.`'.$field.'`) AS `'.$field_name.'`,';
				}
			}
		}
		if ($this->fields_rel_all) {
			foreach ($this->fields_rel_all as $field_name => $fields) {
				foreach ($fields as $table => $field) {
					$result .= ' `'.$table.'`.`'.$field.'` AS `'.$field_name.'`,';
				}
			}
		}
		$result .= ' COUNT(*) AS `count`,';
		if ($this->select_sql) $result .= ' '.$this->select_sql;
		else $result = substr($result, 0, -1).' ';

		return $result;
	}
	
	protected function sql_from($table = NULL)
	{
		if ($table) {
			return ' FROM '.addslashes($table).' ';
		} else {
			return ' FROM '.$this->table.' ';
		}
	}
	
	protected function sql_joins($join_type = 'LEFT')
	{
		$result = ' ';
		if ($this->joins) {
			foreach ($this->joins as $table_field1 => $table_field2) {
				$field_array1 = explode('.',$table_field1);
				$field_array2 = explode('.',$table_field2);
				$table1 = $field_array1[0];
				$table2 = $field_array2[0];
				$field1 = $field_array1[1];
				$field2 = $field_array2[1];
				$result .= ' '.$join_type.' JOIN `'.$table1.'` ON ( `' .
					$table1.'`.`'.$field1.'` = `'.$table2.'`.`'.$field2.'` ) ';
			}
		}
		if ($this->left_joins) {
			foreach ($this->left_joins as $table_field1 => $table_field2) {
				$field_array1 = explode('.',$table_field1);
				$field_array2 = explode('.',$table_field2);
				$table1 = $field_array1[0];
				$table2 = $field_array2[0];
				$field1 = $field_array1[1];
				$field2 = $field_array2[1];
				$result .= ' LEFT JOIN `'.$table1.'` ON ( `' .
					$table1.'`.`'.$field1.'` = `'.$table2.'`.`'.$field2.'` ) ';
			}
		}
		if ($this->right_joins) {
			foreach ($this->right_joins as $table_field1 => $table_field2) {
				$field_array1 = explode('.',$table_field1);
				$field_array2 = explode('.',$table_field2);
				$table1 = $field_array1[0];
				$table2 = $field_array2[0];
				$field1 = $field_array1[1];
				$field2 = $field_array2[1];
				$result .= ' RIGHT JOIN `'.$table1.'` ON ( `' .
					$table1.'`.`'.$field1.'` = `'.$table2.'`.`'.$field2.'` ) ';
			}
		}
		if ($this->joins_sql) $result .= $this->joins_sql;
		return $result;
	}

	protected function sql_group_by($group_by = NULL)
	{
		$result = '';
		if (!$group_by) {
			if ($this->group_by) {
				$group_by = $this->group_by;
			} elseif($this->fields_sum || $this->fields_count) {
				$group_by = $this->key;
			}
		}
		if ($group_by) {
			$groups = explode(',',$group_by);
			$result .= ' GROUP BY';
			foreach ($groups as $group) {
				$group_array = (explode('.',$group));
				if (count($group_array) > 1) {
					$result .= ' `'.$group_array[0].'`.`'.$group_array[1].'`,';
				} else {
					$result .= ' `'.$group_array[0].'`,';
				}
			}
			$result = substr($result, 0, -1).' ';
			return $result;
		} else {
			return ' ';
		}
	}
	
	protected function sql_order_by()
	{
		if ($this->order_by) {
			$orders = explode(',',$this->order_by);
			$result = ' ORDER BY';
			foreach ($orders as $order) {
				$order_array = (explode('.',$order));
				if (count($order_array) > 1) {
					$table = $order_array[0];
					$field = $order_array[1];
				} else {
					$table = $this->table;
					$field = $order_array[0];
				}
				if (substr($field,-1,1) == '-') {
					$field = substr($field,0,-1);
					$direction = 'DESC';
				} else {
					$direction = 'ASC';
				}
				$result .= ' `'.$table.'`.`'.$field.'` '.$direction.',';
			}
			$result = substr($result, 0, -1).' ';
			return $result;
		} else {
			return ' ';
		}
	}

	/**
	 * Creates a WHERE clause to use in an SQL query.
	 * 
	 * Basicly it accepts anything you throw at it. When you provide a string
	 * or integer it will just match that to the dao's key. If you provide a
	 * model it will try to find anything that matches the model.
	 *
	 * @param mixed $model A model or id to match.
	 * @return string
	 */
	protected function sql_where($model = false, $sql = false)
	{
		$result = ' WHERE ';
		if (!$sql) $sql = $this->where;
		$where = false;
		if ($model) {
			if (is_array($model) || is_object($model)) {
				if ($this->foreign_keys) {
					$keys = explode(',',$this->foreign_keys);
					foreach ($keys as $field) {
						if ($model->$field != '' && $model->$field != false) {
							$where = true;
							$values = explode(',',$model->$field);
							foreach ($values as $value) {
								$result .= ' `'.$this->table.'`.`'.$field.'` = "'.addslashes($model->$field).'" OR';
							}
							if ($where) $result = substr($result,0,-2).' AND';
						}
					}
				}
				if ($this->other_keys) {
					$keys = explode(',',$this->other_keys);
					foreach ($keys as $key) {
						$key_array = (explode('.',$key));
						if (count($key_array) > 1) {
							$table = $key_array[0];
							$field = $key_array[1];
						} else {
							$table = $this->table;
							$field = $key_array[0];
						}
						if ($model->$field != '' && $model->$field != false) {
							$where = true;
							$values = explode(',',$model->$field);
							foreach ($values as $value) {
								$result .= ' '.$table.'.'.$field.' = \''.addslashes($model->$field).'\' OR';
							}
							if ($where) $result = substr($result,0,-2).' AND';
						}
					}
				}
				if ($where) $result = substr($result,0,-3).' ';
			} else {
				$values = explode(',',$model);
				$where = true;
				foreach ($values as $value) {
					$result .= ' `'.$this->table.'`.`'.$this->key.'` = "'.addslashes($value).'" OR';
				}
				$result = substr($result,0,-2);
			}
		}

		if ($sql) {
			if ($where) {
				$result .= ' AND ';
			}
			$result .= $sql;
			$where = true;
		}

		if (!$where) {
			$result .= '1 ';
		}
		
		return $result;
	}

	protected function sql_where_like($model = false, $search = false, $keys = false, $sql = false)
	{
		$result = ' WHERE ';
		if (!$sql) $sql = $this->where;
		$where = false;
		if ($model) {
			if ((is_array($model) || is_object($model)) && $this->foreign_keys) {
				$foreign_keys = explode(',',$this->foreign_keys);
				foreach ($foreign_keys as $key) {
					if ($model->$key != '' && $model->$key != false) {
						if (!$where) $result .= '(';
						$where = true;
						$values = explode(',',$model->$key);
						foreach ($values as $value) {
							$result .= ' `'.$this->table.'`.`'.$key.'` = "'.$model->$key.'" OR';
						}
						if ($where) $result = substr($result,0,-2).' AND';
					}
				}
				if ($where) $result = substr($result,0,-3).' ';
			} elseif (!is_array($model) && !is_object($model)) {
				$values = explode(',',$model);
				$where = true;
				foreach ($values as $value) {
					$result .= ' `'.$this->table.'`.`'.$this->key.'` = "'.$value.'" OR';
				}
				$result = substr($result,0,-2);
			}
		}

		if ($search) {
			if ($where) $result .= ' AND ';
			$searches = explode(' ',$search);
			$search = null;

			$search_model = $this->fetch_model();

			foreach ($searches as $search_string) {
				if (ereg('[^ :]*:[^ :]*', $search_string, $regs)) {
					$search_field = explode(':',$regs[0]);
					foreach ($search_model as $key => $value) {
						if ($key == $search_field[0]) {
							if (!$where) $result .= '(';
							$where = true;

							if (substr($search_field[1],0,1) == '!') {
								$search_temp = substr($search_field[1],1);
								$search_span = explode('~',$search_temp);
								if (count($search_span) == 2) {
									$result .= ' !(`'.$this->table.'`.`'.addslashes($search_field[0]).'` >= "'.addslashes($search_span[0]).'"'
										.' AND `'.$this->table.'`.`'.addslashes($search_field[0]).'` <= "'.addslashes($search_span[1]).'") AND';
								} else {
									$result .= ' `'.$this->table.'`.`'.addslashes($search_field[0]).'` != "'.addslashes($search_temp).'" AND';
								}
							} else {
								$search_span = explode('~',$search_field[1]);
								if (count($search_span) == 2) {
									$result .= ' (`'.$this->table.'`.`'.addslashes($search_field[0]).'` >= "'.addslashes($search_span[0]).'"'
										.' AND `'.$this->table.'`.`'.addslashes($search_field[0]).'` <= "'.addslashes($search_span[1]).'") AND';
								} else {
									$result .= ' `'.$this->table.'`.`'.addslashes($search_field[0]).'` = "'.addslashes($search_field[1]).'" AND';
								}
							}
						}
					}
					if (is_array($this->fields_rel)) {
						foreach ($this->fields_rel as $key => $field_array) {
							if ($key == $search_field[0]) {
								if (!$where) $result .= '(';
								$where = true;
	
								$search_span = explode('~',$search_field[1]);
								foreach ($field_array as $table => $field) {
									if (count($search_span) == 2) {
										$result .= ' ('.$table.'.'.$field.' >= \''.addslashes($search_span[0]).'\''
											.' AND '.$table.'.'.$field.' <= \''.addslashes($search_span[1]).'\') AND';
									} else {
										$result .= ' '.$table.'.'.$field.' = \''.addslashes($search_field[1]).'\' AND';
									}
								}
							}
						}
					}
					if (is_array($this->fields_rel_all)) {
						foreach ($this->fields_rel_all as $key => $field_array) {
							if ($key == $search_field[0]) {
								if (!$where) $result .= '(';
								$where = true;
	
								$search_span = explode('~',$search_field[1]);
								foreach ($field_array as $table => $field) {
									if (count($search_span) == 2) {
										$result .= ' ('.$table.'.'.$field.' >= \''.addslashes($search_span[0]).'\''
											.' AND '.$table.'.'.$field.' <= \''.addslashes($search_span[1]).'\') AND';
									} else {
										$result .= ' '.$table.'.'.$field.' = \''.addslashes($search_field[1]).'\' AND';
									}
								}
							}
						}
					}
				} else {
					$search .= $search_string.' ';
				}
			}
			if ($where) $result = substr($result,0,-3).' ';

			if ($search) {
				$search = substr($search,0,-1);
			}

		}
		
		if ($search) {
			if ($where) $result .= ' AND ';
			$keys_m = explode(',',$this->search_fields);
			$keys_l = explode(',',$this->search_fields_l);
			$keys_r = explode(',',$this->search_fields_r);
			foreach ($keys_m as $key) {
				if ($key != '' && $key != false) {
					if (!$where) $result .= '(';
					$where = true;
					$key_array = (explode('.',$key));
					$result .= ' `'.$key_array[0].'`.`'.$key_array[1].'` LIKE "%'.addslashes($search).'%" OR';
				}
			}
			foreach ($keys_l as $key) {
				if ($key != '' && $key != false) {
					if (!$where) $result .= '(';
					$where = true;
					$key_array = (explode('.',$key));
					$result .= ' `'.$key_array[0].'`.`'.$key_array[1].'` LIKE "'.addslashes($search).'%" OR';
				}
			}
			foreach ($keys_r as $key) {
				if ($key != '' && $key != false) {
					if (!$where) $result .= '(';
					$where = true;
					$key_array = (explode('.',$key));
					$result .= ' `'.$key_array[0].'`.`'.$key_array[1].'` LIKE "%'.$search.'" OR';
				}
			}
			if ($where) $result = substr($result,0,-2).' ';
		}

		if ($where) {
			 $result .= ' ) ';
		}
		
		if ($sql) {
			if ($where) {
				$result .= ' AND ';
			}
			$result .= $sql;
			$where = true;
		}

		if (!$where) {
			$result .= '1 ';
		}
		
		return $result;
	}

	protected function sql_limit($limit = NULL, $offset = NULL)
	{
		if ($limit == NULL) $limit = $this->limit;
		if ($offset == NULL) $offset = $this->offset;
		$result = ' ';
		if ($limit) {
			$result .= 'LIMIT ';
			if ($offset) $result .= addslashes($offset).',';
			$result .= addslashes($limit).' ';
		} elseif ($offset) {
			$result .= 'LIMIT '.addslashes($offset).',18446744073709551615 ';
		}
		return $result;
	}
}

?>
